#ifndef _OPERATOR_H
#define _OPERATOR_H

#include "astarte/batch_config.h"
#include "astarte/catype.h"
#include "astarte/machine_view.h"
#include "astarte/parallel_tensor.h"
#include "astarte/utils/dot/record_formatter.h"
#include <vector>

namespace astarte {

extern LegionRuntime::Logger::Category log_measure;

class OpMeta;
class Simulator;
class CostMetrics;

enum class MappingRecordType { INPUT_OUTPUT, INPUT_WEIGHT };

enum class MappingOperation { PARTITION, REPLICATE };

/** @brief A class to keep track of a dimension relation between two tensors
 * used by an operator.
 * 
 * Dimension relations are one-to-one mappings between the dimensions of the
 * input, weights, and output tensors of an operator. Introduced in the Unity
 * paper, dimension relations allow Astarte to keep track of an operator's
 * parallelization plans as part of the Parallel Computation Graph (PCG).
 * 
 * Each ParallelDimMappingRecord only keeps track of a single dimension
 * relation.
 * 
 * ParallelDimMappingRecord objects must be initialized with a MappingRecordType,
 * which can be INPUT_OUTPUT, if the ParallelDimMappingRecord is tracking a dimension
 * relation between the input and the output tensor, or INPUT_WEIGHT, if the 
 * ParallelDimMappingRecord is tracking a dimension relation between the input
 * tensor and the weights tensor.
 * 
*/

class ParallelDimMappingRecord {
private:
    ParallelDimMappingRecord(MappingRecordType);

public:
    /**
     * @brief We disable this constructor because ParallelDimMappingRecord objects
     * must specify the MappingRecordType upon creation.
    */
    ParallelDimMappingRecord() = delete;

    static ParallelDimMappingRecord input_output_record(
        int input_idx,
        int input_dim,
        int output_idx,
        int output_dim,
        tl::optional<MappingOperation> operation = tl::nullopt);
    static ParallelDimMappingRecord input_weight_record(
        int input_idx,
        int input_dim,
        int weight_idx,
        int weight_dim,
        tl::optional<MappingOperation> operation = tl::nullopt);
    MappingRecordType get_type() const;

public:
    MappingRecordType type;
    tl::optional<MappingOperation> operation;

    int output_dim, input_dim, weight_dim;
    int output_idx, input_idx, weight_idx;
};

class Op {
public:
    static void construct_weight_parallel_dims(
        std::vector<ParallelDimMappingRecord> &records,
        std::vector<std::pair<int, int>> mappings,
        int input_idx = 0,
        int weight_idx = 0);
    static void construct_weight_parallel_dims(
        std::vector<ParallelDimMappingRecord> &records,
        std::vector<std::tuple<int, MappingOperation, int>> mappings,
        int input_idx = 0,
        int weight_idx = 0);
    static void construct_weight_parallel_dims(
        std::vector<ParallelDimMappingRecord> &records,
        int input_dim,
        int weight_dim,
        int input_idx = 0,
        int weight_idx = 0,
        tl::optional<MappingOperation> operation = tl::nullopt);
    static void construct_output_parallel_dims(
        std::vector<ParallelDimMappingRecord> &records,
        std::vector<std::pair<int, int>> mappings,
        int input_idx = 0,
        int output_idx = 0);
    static void construct_output_parallel_dims(
        std::vector<ParallelDimMappingRecord> &records,
        std::vector<std::tuple<int, MappingOperation, int>> mappings,
        int input_idx = 0,
        int output_idx = 0);
    static void construct_output_parallel_dims(
        std::vector<ParallelDimMappingRecord> &records,
        int input_dim,
        int output_dim,
        int input_idx = 0,
        int output_idx = 0,
        tl::optional<MappingOperation> operation = tl::nullopt);

    ParallelConfig view_to_pc(MachineView const &view) const;

protected:
    void register_weight_parallel_dims(
        std::vector<std::tuple<int, MappingOperation, int>> mappings,
        int input_idx = 0,
        int weight_idx = 0);
    void register_weight_parallel_dims(
        int input_dim,
        int weight_dim,
        int input_idx = 0,
        int weight_idx = 0,
        tl::optional<MappingOperation> operation = tl::nullopt);
    void register_output_parallel_dims(
        std::vector<std::pair<int, int>> mappings,
        int input_idx = 0,
        int output_idx = 0);
    void register_output_parallel_dims(
        std::vector<std::tuple<int, MappingOperation, int>> mappings,
        int input_idx = 0,
        int output_idx = 0);
    void register_output_parallel_dims(
        int input_dim,
        int output_dim,
        int input_idx = 0,
        int output_idx = 0,
        tl::optional<MappingOperation> operation = tl::nullopt);
    
    int get_output_to_input_dim_mapping(const ParallelTensor output,
                                        int output_dim,
                                        const ParallelTensor input);
    int get_output_to_weight_dim_mapping(const ParallelTensor output,
                                         int output_dim,
                                         const ParallelTensor weight);
    
    void inner_measure_operator_cost(Simulator *sim,
                                     std::function<void()> const &forward,
                                     std::function<void()> const &backward,
                                     CostMetrics &cost_metrics) comparison_fn_t
    bool check_output_input_weight_parallel_dims(
        bool allocate_weights = true) const;
    bool check_output_input_weight_same_parallel_is() const;
    bool check_output_input_weight_same_machine_view() const;

public:
  Op(CAModel &model,
     OperatorType otype,
     DataType dtype,
     char const *_name,
     int numInputs,
     int numWeights,
     bool allocate_weights,
     int numOutputs,
     const ParallelTensor input1 = NULL,
     const ParallelTensor input2 = NULL,
     const ParallelTensor input3 = NULL,
     const ParallelTensor input4 = NULL);
  Op(CAModel &model,
     OperatorType otype,
     DataType dtype,
     char const *_name,
     int numInputs,
     int numWeights,
     int numOutputs,
     const ParallelTensor input1 = NULL,
     const ParallelTensor input2 = NULL,
     const ParallelTensor input3 = NULL,
     const ParallelTensor input4 = NULL);
  Op(int guid,
     bool profiling,
     OperatorType otype,
     DataType dtype,
     char const *name,
     int numInputs,
     int numWeights,
     int numOutputs,
     const ParallelTensor input1 = NULL,
     const ParallelTensor input2 = NULL,
     const ParallelTensor input3 = NULL,
     const ParallelTensor input4 = NULL);
  Op(CAModel &model,
     OperatorType otype,
     DataType dtype,
     char const *_name,
     int numInputs,
     int numWeights,
     int numOutputs,
     ParallelTensor const *tensors);
  // graph substitution related methods
  virtual bool get_int_parameter(PMParameter, int *) const;
  virtual bool get_tensor_parameter(TNParameter, DIMParameter, int *) const;
  virtual bool get_input_parameter(TNParameter, DIMParameter, int *) const;
  virtual bool get_weight_parameter(TNParameter, DIMParameter, int *) const;
  // Pure virtual functions that must be implemented
  virtual void init(CAModel const &) = 0;
  virtual void init_inference(CAModel const &,
                              std::vector<ParallelTensor> const &,
                              std::vector<ParallelTensor> const &,
                              MachineView const *mv = nullptr) {
    assert(false);
  };
  virtual void forward(CAModel const &) = 0;
  virtual void backward(CAModel const &) = 0;
  // Pure virtual functions for inference
  virtual Legion::FutureMap inference(CAModel const &,
                                      BatchConfigFuture const &,
                                      std::vector<ParallelTensor> const &,
                                      std::vector<ParallelTensor> const &,
                                      MachineView const *mv = nullptr) {
    assert(false);
  };
  virtual void print_layer(CAModel const &model) = 0;
  virtual bool measure_operator_cost(Simulator *sim,
                                     MachineView const &mv,
                                     CostMetrics &cost_metrics) const = 0;
  virtual bool estimate_sync_cost(Simulator *sim,
                                  MachineView const &pc,
                                  CostMetrics &cost_metrics) const;
  // Other virtual functions that can be optionally overwritten
  virtual ParallelConfig get_random_parallel_config(CAModel const &ff) const;
  virtual ParallelConfig get_data_parallel_config(CAModel const &ff) const;
  virtual Legion::Domain get_input_tensor_shape(ParallelConfig const &pc,
                                                int input_idx,
                                                int part_idx) const;
  virtual Legion::Domain get_output_tensor_shape(ParallelConfig const &pc,
                                                 int output_idx,
                                                 int part_idx) const;
  virtual Legion::Domain get_weight_tensor_shape(ParallelConfig const &pc,
                                                 int weight_idx,
                                                 int part_idx) const;
  virtual bool is_valid_parallel_config(CAModel const &ff,
                                        ParallelConfig const &pc) const;
  virtual bool is_adoptable_parallel_config(CAModel const &ff,
                                            ParallelConfig const &pc) const;
  // Helper functions
  void prefetch(CAModel const &);
  void zero_grad(CAModel const &);
  ParallelTensor get_parameter(int index);
  virtual void map_output_tensors(CAModel &ff);
  virtual bool can_inplace_output();
  virtual bool has_inplace_output();
  virtual void do_inplace_output();
  virtual bool is_parallel_op() const;
  virtual void serialize(Legion::Serializer &) const;
  virtual Op *
      materialize(CAModel &ff, ParallelTensor inputs[], int num_inputs) const;
  size_t get_untyped_params_hash() const;
  virtual size_t get_params_hash() const;

  virtual tl::optional<RecordFormatter> as_dot() const;

  int get_dimension() const;
#ifdef CA_USE_NCCL
  static ncclUniqueId get_nccl_unique_id_task(
      Legion::Task const *task,
      std::vector<Legion::PhysicalRegion> const &regions,
      Legion::Context ctx,
      Legion::Runtime *runtime);
  static ncclComm_t
      init_nccl_comms_task(Legion::Task const *task,
                           std::vector<Legion::PhysicalRegion> const &regions,
                           Legion::Context ctx,
                           Legion::Runtime *runtime);
#endif
protected:
  void set_argumentmap_for_init(CAModel const &ff, Legion::ArgumentMap &argmap);
  void set_argumentmap_for_init_inference(CAModel const &ff,
                                          Legion::ArgumentMap &argmap,
                                          ParallelTensor const output0);
  void set_argumentmap_for_forward(CAModel const &ff,
                                   Legion::ArgumentMap &argmap);
  void set_argumentmap_for_inference(CAModel const &ff,
                                     Legion::ArgumentMap &argmap,
                                     ParallelTensor const output0);
  void set_argumentmap_for_backward(CAModel const &ff,
                                    Legion::ArgumentMap &argmap);
  void set_opmeta_from_futuremap(CAModel const &ff,
                                 Legion::FutureMap const &fm);
  void set_opmeta_from_futuremap_inference(CAModel const &ff,
                                           Legion::FutureMap const &fm,
                                           ParallelTensor const output0);
  void solve_parallel_dim_mappings(
      std::vector<ParallelDim const *> const &inputs,
      std::vector<ParallelDim *> const &weights,
      std::vector<ParallelDim *> const &outputs) const;

public:
  OperatorType op_type;
  DataType data_type;
  // the guid of the layer associated with the current operator
  // layer_guid is used to match layer with op
  LayerID layer_guid;
  size_t op_guid;
  char name[MAX_OPNAME];
  Legion::IndexSpace parallel_is;
  ParallelTensor outputs[MAX_NUM_OUTPUTS];
  ParallelTensor inputs[MAX_NUM_INPUTS];
  ParallelParameter weights[MAX_NUM_WEIGHTS];
  bool trainableInputs[MAX_NUM_INPUTS];
  OpMeta *meta[MAX_NUM_WORKERS];
  std::map<ParallelTensor, OpMeta *[MAX_NUM_WORKERS]> inference_meta;
  int numInputs, numWeights, numOutputs;
  bool profiling;
  bool add_bias_only_once;
#ifdef CA_USE_NCCL
  ncclUniqueId ncclId;
#endif
  // Note: parallel_dims_mapping should not be called in a DNN task
  std::vector<ParallelDimMappingRecord> *parallel_dims_mapping;
};


}; // namespace astarte

#endif // _OPERATOR_H